from json_database import JsonStorageXDG
from os.path import join, exists, isdir
from os import makedirs
from jarbas_hive_mind.settings import DEFAULT_PORT, DATA_PATH, CERTS_PATH, \
    MYCROFT_WEBSOCKET_CONFIG


def default_config():
    return {'data_path': DATA_PATH,
            'database': join(DATA_PATH, "database", "clients.db"),
            'log_blacklist': [],
            'max_connections': -1,
            'mycroft_bus': MYCROFT_WEBSOCKET_CONFIG,
            'port': DEFAULT_PORT,
            'ssl': {'certificates': CERTS_PATH,
                    'ssl_certfile': 'HiveMind.crt',
                    'ssl_keyfile': 'HiveMind.key'}
            }


def _merge_defaults(base, default=None):
    """
        Recursively merging configuration dictionaries.

        Args:
            base:  Target for merge
            default: Dictionary to merge into base if key not present
    """
    default = default or default_config()
    for k, dv in default.items():
        bv = base.get(k)
        if isinstance(dv, dict) and isinstance(bv, dict):
            _merge_defaults(bv, dv)
        elif k not in base:
            base[k] = dv
    return base


CONFIGURATION = JsonStorageXDG("HivemindCore")
CONFIGURATION = _merge_defaults(CONFIGURATION)

# ensure directories exist
if not exists(CONFIGURATION.path):
    CONFIGURATION.store()

if not isdir(CONFIGURATION["data_path"]):
    makedirs(CONFIGURATION["data_path"])

if not isdir(CONFIGURATION["ssl"]["certificates"]):
    makedirs(CONFIGURATION["ssl"]["certificates"])



from jarbas_hive_mind.configuration import CONFIGURATION, default_config
from ovos_utils.log import LOG


if "sql" not in CONFIGURATION["database"].split(":/")[0]:
    from jarbas_hive_mind.database.json_db import JsonClientDatabase
    ClientDatabase = JsonClientDatabase
else:
    try:
        from jarbas_hive_mind.database.sql import SQLClientDatabase
        ClientDatabase = SQLClientDatabase
    except ImportError:
        LOG.error("Run pip install sqlalchemy")

        LOG.info("Falling back to json database")
        CONFIGURATION["database"] = default_config()["database"]
        CONFIGURATION.store()
        from jarbas_hive_mind.database.json_db import JsonClientDatabase

        ClientDatabase = JsonClientDatabase




from ovos_utils.log import LOG
from autobahn.twisted.websocket import WebSocketClientFactory, \
    WebSocketClientProtocol
from twisted.internet.protocol import ReconnectingClientFactory
from jarbas_hive_mind.exceptions import UnauthorizedKeyError, \
    SecureConnectionFailed, ConnectionError, HiveMindEntryPointNotFound
from jarbas_hive_mind.utils import encrypt_as_json, decrypt_from_json
from jarbas_hive_mind.interface import HiveMindSlaveInterface
from ovos_utils.messagebus import Message
import json
from twisted.internet import reactor

platform = "HiveMindTerminalv0.2"


class HiveMindTerminalProtocol(WebSocketClientProtocol):

    def onConnect(self, response):
        LOG.info("HiveMind connected: {0}".format(response.peer))
        self.factory.client = self
        self.factory.status = "connected"

    def onOpen(self):
        LOG.info("HiveMind websocket connection open. ")

    def onMessage(self, payload, isBinary):
        if not isBinary:
            payload = self.decode(payload)
            data = {"payload": payload, "isBinary": isBinary}
        else:
            data = {"payload": None, "isBinary": isBinary}
        self.factory.handle_incoming_message(data)

    def onClose(self, wasClean, code, reason):
        LOG.warning("HiveMind websocket connection closed: {0}".format(reason))
        self.factory.client = None
        self.factory.status = "disconnected"
        if "WebSocket connection upgrade failed" in reason:
            # key rejected
            LOG.error("Key rejected")
            #raise UnauthorizedKeyError

        elif self.factory.connection.is_secure:
            if "WebSocket opening handshake timeout" in reason:
                LOG.error("SecureConnectionFailed: " + reason)
                # raise SecureConnectionFailed

        else:
            LOG.error("ConnectionError")
        reactor.stop()

    def decode(self, payload):
        payload = payload.decode("utf-8")
        if self.factory.crypto_key:
            payload = decrypt_from_json(self.factory.crypto_key, payload)
        return payload

    def sendMessage(self,
                    payload,
                    isBinary=False,
                    fragmentSize=None,
                    sync=False,
                    doNotCompress=False):
        if self.factory.crypto_key and not isBinary:
            payload = encrypt_as_json(self.factory.crypto_key, payload)
        if isinstance(payload, dict):
            payload = json.dumps(payload)
        if isinstance(payload, str):
            payload = bytes(payload, encoding="utf-8")
        super().sendMessage(payload, isBinary, fragmentSize=fragmentSize,
                            sync=sync, doNotCompress=doNotCompress)


class HiveMindTerminal(WebSocketClientFactory, ReconnectingClientFactory):
    protocol = HiveMindTerminalProtocol
    announce = False

    def __init__(self, crypto_key=None, connection=None,
                 auto_reconnect=False, *args, **kwargs):
        super(HiveMindTerminal, self).__init__(*args, **kwargs)
        self.status = "disconnected"
        self.client = None
        self.crypto_key = crypto_key
        self.connection = None
        self.interface = HiveMindSlaveInterface(self)
        self.auto_reconnect = auto_reconnect
        self.upnp_server = None
        self.ssdp = None
        if connection:
            self.bind(connection)

    def bind(self, connection):
        self.connection = connection

    @property
    def node_id(self):
        if not self.peer:
            return None
        return self.peer + ":SLAVE"

    @property
    def peer(self):
        if self.client:
            return self.client.peer
        return None

    @property
    def master(self):
        return self.connection.peer

    # parsed protocol messages
    def handle_incoming_mycroft(self, message):
        assert isinstance(message, Message)
        LOG.debug("[Mycroft Message] " + message.serialize())

    # HiveMind protocol messages - from UPstream
    def handle_bus_message(self, payload):
        # Generate mycroft Message
        if isinstance(payload, str):
            payload = json.loads(payload)
        msg_type = payload.get("msg_type") or payload["type"]
        data = payload.get("data") or {}
        context = payload.get("context") or {}
        message = Message(msg_type, data, context)
        message.context["source"] = self.client.peer
        message.context["destination"] = "skills"
        if "platform" not in message.context:
            message.context["platform"] = platform
        self.handle_incoming_mycroft(message)

    def handle_propagate_message(self, payload, msg_data):
        LOG.info("Received propagate message at: " + self.node_id)
        LOG.debug("ROUTE: " + str(msg_data["route"]))
        LOG.debug("PAYLOAD: " + str(payload))
        self.interface.propagate(payload, msg_data)

    def handle_broadcast_message(self, payload, msg_data):
        LOG.info("Received broadcast message at: " + self.node_id)
        LOG.debug("ROUTE: " + str(msg_data["route"]))
        LOG.debug("PAYLOAD: " + str(payload))
        self.interface.broadcast(payload, msg_data)

    def handle_escalate_message(self, payload, msg_data):
        # only Slaves are allowed to escalate, by definition escalate
        # goes upstream only
        LOG.debug("Ignoring escalate message from upstream, illegal action")

    def handle_incoming_message(self, data):
        payload = data.get("payload")
        if data.get("isBinary"):
            self.on_binary(payload)
        else:
            self.on_message(payload)

    # websocket handlers
    def send_to_hivemind_bus(self, payload):
        if isinstance(payload, Message):
            payload = payload.serialize()
        self.interface.send_to_hivemind_bus(payload)
        return payload

    def sendBinary(self, payload):
        if self.client is None:
            LOG.error("Client is none")
            return
        self.client.sendMessage(payload, isBinary=True)

    def sendMessage(self, payload):
        if self.client is None:
            LOG.error("Client is none")
            return
        self.client.sendMessage(payload, isBinary=False)

    def on_binary(self, payload):
        LOG.info("[BINARY MESSAGE]")

    def on_message(self, payload):
        msg = json.loads(payload)
        msg_type = msg["msg_type"]
        payload = msg["payload"]

        # Parse hive protocol
        if msg_type == "bus":
            self.handle_bus_message(payload)
        elif msg_type == "propagate":
            self.handle_propagate_message(payload, msg)
        elif msg_type == "broadcast":
            self.handle_broadcast_message(payload, msg)
        elif msg_type == "escalate":
            self.handle_escalate_message(payload, msg)
        else:
            LOG.error("Unknown HiveMind protocol msg_type")

    def clientConnectionFailed(self, connector, reason):
        self.status = "disconnected"
        if "DNS lookup failed:" in str(reason):
            LOG.error("Could not find the specified HiveMind entry point")
            LOG.debug("Does this look like a valid address? " +
                      self.connection.address)
            raise HiveMindEntryPointNotFound
        else:

            if self.auto_reconnect:
                LOG.error("HiveMind client failed: " + str(reason) +
                          " .. retrying ..")
                self.retry(connector)
            else:
                LOG.error("HiveMind client failed: " + str(reason))

    def clientConnectionLost(self, connector, reason):
        LOG.error("HiveMind connection lost: " + str(reason) +
                  " .. retrying ..")
        self.status = "disconnected"
        if self.auto_reconnect:
            LOG.error("HiveMind connection lost: " + str(reason) +
                      " .. retrying ..")
            self.retry(connector)
        else:
            LOG.error("HiveMind connection lost: " + str(reason))




from jarbas_hive_mind import get_listener, get_connection
from jarbas_hive_mind.slave import HiveMindSlave
from jarbas_hive_mind.configuration import CONFIGURATION
from twisted.internet import reactor


class FakeBus:
    def __init__(self, *args, **kwargs):
        self.events = {}

    def on(self, msg_type, handler):
        if msg_type not in self.events:
            self.events[msg_type] = []
        self.events[msg_type].append(handler)

    def emit(self, message):
        if message.msg_type in self.events:
            for handler in self.events[message.msg_type]:
                handler(message)

    def remove(self, msg_type, handler):
        pass


class FakeMycroft:

    def __init__(self, port, key="dummy_key", config=None):
        self.bus = FakeBus()
        self.config = config or CONFIGURATION
        self.port = port
        self.user_agent = "fakeCroft:{n}".format(n=self.port)
        self.key = key
        # hive mind objects
        self.hive = None
        self.slave_connection = None
        self.connection = None
        self.listener = None

    @property
    def interface(self):
        if self.hive_interface:
            return self.hive_interface
        elif self.slave_interface:
            return self.slave_interface
        else:
            return None

    @property
    def hive_interface(self):
        # the master can send requests downstream with this interface
        if self.hive:
            return self.hive.interface
        return None

    @property
    def slave_interface(self):
        # the slave_connection can send requests upstream with this interface
        if self.slave_connection:
            return self.slave_connection.interface
        return None

    def start_listener(self):
        # listen
        self.listener = get_listener(port=self.port,
                                     bus=self.bus)

        # this flag is not meant for the end user
        # used here so the twisted reactor can be started manually
        self.listener._autorun = False

        # returns a HiveMind object
        self.hive = self.listener.listen()

    def connect(self, port, host="0.0.0.0"):
        self.connection = get_connection(host, port)

        # this flag is not meant for the end user
        # used here so the twisted reactor can be started manually
        self.connection._autorun = False

        con = HiveMindSlave(bus=self.bus,
                            headers=self.connection.get_headers(self.user_agent,
                                                                self.key),
                            useragent=self.user_agent)

        # returns a HiveMindSlave object
        self.slave_connection = self.connection.connect(con)

    def run(self):
        reactor.run()




import json
# If sqlalchemy not installed will fallback to json based database
from sqlalchemy import Column, Text, String, Integer, create_engine, Boolean
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base


from ovos_utils.log import LOG
from jarbas_hive_mind.configuration import CONFIGURATION

Base = declarative_base()


def model_to_dict(obj):
    serialized_data = {c.key: getattr(obj, c.key) for c in obj.__table__.columns}
    return serialized_data


def props(cls):
    return [i for i in cls.__dict__.keys() if i[:1] != '_']


class Client(Base):
    __tablename__ = "clients"
    client_id = Column(Integer, primary_key=True)
    description = Column(Text)
    api_key = Column(String)
    name = Column(String)
    crypto_key = Column(String)
    mail = Column(String)
    last_seen = Column(Integer, default=0)
    is_admin = Column(Boolean, default=False)
    blacklist = Column(Text)  # json string


class SQLClientDatabase:
    def __init__(self, path=CONFIGURATION["database"], debug=False, session=None):
        self.db = create_engine(path)
        self.db.echo = debug
        if session:
            self.session = session
        else:
            Session = sessionmaker(bind=self.db)
            self.session = Session()
        Base.metadata.create_all(self.db)

    def update_timestamp(self, key, timestamp):
        user = self.get_client_by_api_key(key)
        if not user:
            return False
        user.last_seen = timestamp
        return True

    def delete_client(self, key):
        user = self.get_client_by_api_key(key)
        if user:
            self.session.delete(user)
            return True
        return False

    def change_key(self, old_key, new_key):
        user = self.get_client_by_api_key(old_key)
        if not user:
            return False
        user.api_key = new_key
        return True

    def change_crypto_key(self, api_key, new_key):
        user = self.get_client_by_api_key(api_key)
        if not user:
            return False
        user.crypto_key = new_key
        return True

    def change_name(self, new_name, key):
        user = self.get_client_by_api_key(key)
        if not user:
            return False
        user.name = new_name
        return True

    def change_blacklist(self, blacklist, key):
        if isinstance(blacklist, dict):
            blacklist = json.dumps(blacklist)
        user = self.get_client_by_api_key(key)
        if not user:
            return False
        user.blacklist = blacklist
        return True

    def get_blacklist_by_api_key(self, api_key):
        user = self.session.query(Client).filter_by(api_key=api_key).first()
        return json.loads(user.blacklist)

    def get_client_by_api_key(self, api_key):
        return self.session.query(Client).filter_by(api_key=api_key).first()

    def get_client_by_name(self, name):
        return self.session.query(Client).filter_by(name=name).all()

    def get_crypto_key(self, api_key):
        user = self.get_client_by_api_key(api_key)
        if not user:
            return None
        return user.crypto_key

    def add_client(self, name=None, mail=None, key="", admin=False,
                   blacklist="{}", crypto_key="RESISTENCEisFUTILE"):
        if isinstance(blacklist, dict):
            blacklist = json.dumps(blacklist)

        user = self.get_client_by_api_key(key)
        if crypto_key is not None:
            crypto_key = crypto_key[:16]
        if user:
            user.name = name
            user.mail = mail
            user.blacklist = blacklist
            user.is_admin = admin
            user.crypto_key = crypto_key if crypto_key else user.crypto_key
        else:
            user = Client(api_key=key, name=name, mail=mail,
                          blacklist=blacklist, crypto_key=crypto_key,
                          client_id=self.total_clients() + 1,
                          is_admin=admin)
            self.session.add(user)

    def total_clients(self):
        return self.session.query(Client).count()

    def commit(self):
        try:
            self.session.commit()
            return True
        except IntegrityError:
            self.session.rollback()
        return False

    def close(self):
        self.session.close()

    def __enter__(self):
        """ Context handler """
        return self

    def __exit__(self, _type, value, traceback):
        """ Commits changes and Closes the session """
        try:
            self.commit()
        except Exception as e:
            LOG.error(e)
        finally:
            self.close()




import base64
from twisted.internet import reactor, ssl
from twisted.internet.error import ReactorNotRunning
from jarbas_hive_mind.master import HiveMind, HiveMindProtocol
from jarbas_hive_mind.configuration import CONFIGURATION
from jarbas_hive_mind.settings import DEFAULT_PORT
from jarbas_hive_mind.utils import create_self_signed_cert
from jarbas_hive_mind.exceptions import SecureConnectionFailed, ConnectionError
from ovos_utils.messagebus import get_mycroft_bus
from ovos_utils.log import LOG
from os.path import join, exists, isfile
from twisted.python.log import err
import logging

logging.getLogger("urllib3.connectionpool").setLevel("INFO")
logging.getLogger("Service").setLevel("INFO")
logging.getLogger("Device").setLevel("INFO")
logging.getLogger("JsonDatabase").setLevel("INFO")


class HiveMindConnection:
    _autorun = True

    def __init__(self, host="127.0.0.1", port=DEFAULT_PORT):
        host = host.replace("https://", "wss://").replace("http://", "ws://")
        if "wss://" in host:
            secure = True
        else:
            secure = False
        self._secure = secure

        host = host.replace("wss://", "").replace("ws://", "")
        self.host = host
        self.port = port

    @property
    def is_secure(self):
        return self._secure

    @property
    def address(self):
        if self.is_secure:
            if "wss://" in self.host:
                return self.host + u":" + str(self.port)
            return "wss://" + self.host + u":" + str(self.port)
        else:
            if "ws://" in self.host:
                return self.host + u":" + str(self.port)
            return "ws://" + self.host + u":" + str(self.port)

    @property
    def peer(self):
        return "tcp4:" + self.host + ":" + str(self.port)

    @staticmethod
    def get_headers(name, key):
        # Note that keys can be shared across users
        # name is not used for auth
        name = name.replace(":", "__")
        authorization = bytes(name + ":" + key, encoding="utf-8")
        key = base64.b64encode(authorization)
        headers = {'authorization': key}
        return headers

    def secure_connect(self, component):
        self._secure = True
        component.bind(self)
        LOG.info("Connecting securely to " + self.address)
        contextFactory = ssl.ClientContextFactory()
        self._run(reactor.connectSSL(self.host, self.port, component, contextFactory))
        return component

    def unsafe_connect(self, component):
        self._secure = False
        component.bind(self)
        LOG.info("Connecting to " + self.address)
        LOG.warning("This listener is unsecured")
        self._run(reactor.connectTCP(self.host, self.port, component))
        return component

    def _run(self, component):

        def fatalError(reason):
            print(reason)
            err(reason, "Absolutely needed the foo, could not get it")
            reactor.stop()

       # component.addErrback(fatalError)

        if not reactor.running and self._autorun:
            try:
                reactor.run()
            except Exception as e:
                print(e)

    def connect(self, component):
        try:
            if self.is_secure:
                try:
                    return self.secure_connect(component)
                except ConnectionError:
                    raise SecureConnectionFailed
            else:
                return self.unsafe_connect(component)
        except Exception as e:
            print(e)
            raise e


class HiveMindListener:
    _autorun = True

    def __init__(self, port=DEFAULT_PORT, max_cons=-1, bus=None):
        self.host = "0.0.0.0"
        self.port = port
        self.max_cons = max_cons
        self._use_ssl = CONFIGURATION["ssl"].get("use_ssl", False)
        self.certificate_path = CONFIGURATION["ssl"]["certificates"]
        self._key_file = CONFIGURATION["ssl"].get("ssl_keyfile",
                                                  "HiveMind.key")
        self._cert_file = CONFIGURATION["ssl"].get("ssl_certfile",
                                                   "HiveMind.crt")
        self.bus = bus

    def bind(self, bus=None):
        # TODO read config for bus options
        self.bus = bus or get_mycroft_bus()

    @property
    def ssl_key(self):
        if isfile(self._key_file):
            return self._key_file
        return join(self.certificate_path, self._key_file)

    @property
    def ssl_cert(self):
        if isfile(self._cert_file):
            return self._cert_file
        return join(self.certificate_path, self._cert_file)

    def load_config(self, config=CONFIGURATION, gen_keys=True):
        # read configuration
        self.port = config["port"]
        self.max_cons = config.get("max_connections", -1)

        ssl_config = config.get("ssl", {})
        self.certificate_path = ssl_config.get("certificates",
                                               self.certificate_path)
        self._key_file = ssl_config.get("ssl_keyfile", self._key_file)
        self._cert_file = ssl_config.get("ssl_certfile", self._cert_file)
        self._use_ssl = ssl_config.get("use_ssl", True)

        # generate self signed keys
        if not exists(self.ssl_key) and gen_keys and \
                self.is_secure:
            LOG.warning("ssl keys dont exist")
            self.generate_keys(self.certificate_path)

    @property
    def is_secure(self):
        return self._use_ssl

    @property
    def address(self):
        if self.is_secure:
            return "wss://" + self.host + ":" + str(self.port)
        return "ws://" + self.host + ":" + str(self.port)

    @property
    def peer(self):
        return "tcp4:" + self.host + ":" + str(self.port)

    @staticmethod
    def generate_keys(path=CONFIGURATION["ssl"]["certificates"],
                      key_name="HiveMind"):
        LOG.info("creating self signed SSL keys")
        name = key_name.split("/")[-1].replace(".key", "")
        create_self_signed_cert(path, name)
        cert = path + "/" + name + ".crt"
        key = path + "/" + name + ".key"
        LOG.info("key created at: " + key)
        LOG.info("crt created at: " + cert)

    def secure_listen(self, key=None, cert=None, factory=None, protocol=None):
        self._use_ssl = True
        key = key or self.ssl_key
        cert = cert or self.ssl_cert

        # SSL server context: load server key and certificate
        contextFactory = ssl.DefaultOpenSSLContextFactory(key, cert)

        factory = factory or HiveMind(bus=self.bus)
        factory.protocol = protocol or HiveMindProtocol
        if self.max_cons >= 0:
            factory.setProtocolOptions(maxConnections=self.max_cons)
        factory.bind(self)
        self.factory = factory
        reactor.listenSSL(self.port, factory, contextFactory)
        LOG.info("HiveMind Listening: " + self.address)
        if self._autorun and not reactor.running:
            reactor.run()
        return factory

    def unsafe_listen(self, factory=None, protocol=None):
        self._use_ssl = False
        factory = factory or HiveMind(bus=self.bus)
        factory.protocol = protocol or HiveMindProtocol
        if self.max_cons >= 0:
            factory.setProtocolOptions(maxConnections=self.max_cons)
        factory.bind(self)
        self.factory = factory
        reactor.listenTCP(self.port, factory)
        LOG.info("HiveMind Listening (UNSECURED): " + self.address)
        if self._autorun and not reactor.running:
            reactor.run()
        return factory

    def listen(self, factory=None, protocol=None):
        if self.is_secure:
            return self.secure_listen(factory=factory, protocol=protocol)
        else:
            return self.unsafe_listen(factory=factory, protocol=protocol)

    def stop(self):
        """Stop the reactor and join the reactor thread until it stops.
        """
        try:
            reactor.stop()
        except ReactorNotRunning:
            LOG.info("twisted reactor stopped")
        except Exception as e:
            LOG.error(e)

    def stop_from_thread(self):
        reactor.callFromThread(self.stop)
        for p in reactor.getDelayedCalls():
            if p.active():
                p.cancel()


def get_listener(port=DEFAULT_PORT, max_connections=-1, bus=None):
    return HiveMindListener(port, max_connections, bus)


def get_connection(host="127.0.0.1", port=DEFAULT_PORT):
    return HiveMindConnection(host, port)



import logging
import random
from os import makedirs
from os.path import exists, join
from socket import gethostname

from OpenSSL import crypto
import json
from jarbas_hive_mind.settings import CERTS_PATH, LOG_BLACKLIST
from jarbas_hive_mind.exceptions import DecryptionKeyError, EncryptionKeyError
from ovos_utils.log import LOG
from ovos_utils.security import encrypt, decrypt
from binascii import hexlify, unhexlify

# this used to be a method here, keep here for now in case something is
# importing it, TODO deprecate
from ovos_utils import get_ip


def validate_param(value, name):
    if not value:
        raise ValueError("Missing or empty %s in conf " % name)


def create_self_signed_cert(cert_dir=CERTS_PATH, name="jarbas_hivemind"):
    """
    If name.crt and name.key don't exist in cert_dir, create a new
    self-signed cert and key pair and write them into that directory.
    """

    CERT_FILE = name + ".crt"
    KEY_FILE = name + ".key"
    cert_path = join(cert_dir, CERT_FILE)
    key_path = join(cert_dir, KEY_FILE)

    if not exists(join(cert_dir, CERT_FILE)) \
            or not exists(join(cert_dir, KEY_FILE)):
        # create a key pair
        k = crypto.PKey()
        k.generate_key(crypto.TYPE_RSA, 1024)

        # create a self-signed cert
        cert = crypto.X509()
        cert.get_subject().C = "PT"
        cert.get_subject().ST = "Europe"
        cert.get_subject().L = "Mountains"
        cert.get_subject().O = "Jarbas AI"
        cert.get_subject().OU = "Powered by Mycroft-Core"
        cert.get_subject().CN = gethostname()
        cert.set_serial_number(random.randint(0, 2000))
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(10 * 365 * 24 * 60 * 60)
        cert.set_issuer(cert.get_subject())
        cert.set_pubkey(k)
        cert.sign(k, 'sha1')
        if not exists(cert_dir):
            makedirs(cert_dir)
        open(cert_path, "wb").write(
            crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
        open(join(cert_dir, KEY_FILE), "wb").write(
            crypto.dump_privatekey(crypto.FILETYPE_PEM, k))

    return cert_path, key_path


def create_echo_function(name, whitelist=None):
    """ Standard logging mechanism for Mycroft processes.

    This handles the setup of the basic logging for all Mycroft
    messagebus-based processes.

    Args:
        name (str): Reference name of the process
        whitelist (list, optional): List of "type" strings.  If defined, only
                                    messages in this list will be logged.

    Returns:
        func: The echo function
    """
    blacklist = LOG_BLACKLIST

    def echo(message):
        global _log_all_bus_messages
        try:
            msg = json.loads(message)

            if whitelist and msg.get("type") not in whitelist:
                return

            if blacklist and msg.get("type") in blacklist:
                return

            if msg.get("type") == "mycroft.debug.log":
                # Respond to requests to adjust the logger settings
                lvl = msg["data"].get("level", "").upper()
                if lvl in ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]:
                    LOG.level = lvl
                    LOG(name).info("Changing log level to: {}".format(lvl))
                    try:
                        logging.getLogger('urllib3').setLevel(lvl)
                    except Exception:
                        pass  # We don't really care about if this fails...

                # Allow enable/disable of messagebus traffic
                log_bus = msg["data"].get("bus", None)
                if log_bus is not None:
                    LOG(name).info("Bus logging: " + str(log_bus))
                    _log_all_bus_messages = log_bus
            elif msg.get("type") == "registration":
                # do not log tokens from registration messages
                msg["data"]["token"] = None
                message = json.dumps(msg)
        except Exception:
            pass

        if _log_all_bus_messages:
            # Listen for messages and echo them for logging
            LOG(name).debug(message)

    return echo


def serialize_message(message):
    # convert a Message object into raw data that can be sent over
    # websocket
    if hasattr(message, 'serialize'):
        return message.serialize()
    else:
        return json.dumps(message.__dict__)


def encrypt_as_json(key, data, nonce=None):
    if isinstance(data, dict):
        data = json.dumps(data)
    if len(key) > 16:
        key = key[0:16]
    try:
        ciphertext, tag, nonce = encrypt(key, data, nonce=nonce)
    except:
        raise EncryptionKeyError
    return json.dumps({"ciphertext": hexlify(ciphertext).decode('utf-8'),
            "tag": hexlify(tag).decode('utf-8'),
            "nonce": hexlify(nonce).decode('utf-8')})


def decrypt_from_json(key, data):
    if isinstance(data, str):
        data = json.loads(data)
    if len(key) > 16:
        key = key[0:16]
    ciphertext = unhexlify(data["ciphertext"])
    if data.get("tag") is None:  # web crypto
        ciphertext, tag = ciphertext[:-16], ciphertext[-16:]
    else:
        tag = unhexlify(data["tag"])
    nonce = unhexlify(data["nonce"])
    try:
        return decrypt(key, ciphertext, tag, nonce)
    except ValueError:
        raise DecryptionKeyError



import json
from jarbas_hive_mind.configuration import CONFIGURATION
from json_database import JsonDatabase
from ovos_utils.log import LOG


class Client:
    def __init__(self, client_id, api_key, name="", mail="",
                 description="", is_admin=False, last_seen=-1,
                 blacklist=None, crypto_key=None):
        self.client_id = client_id
        self.description = description
        self.api_key = api_key
        self.name = name
        self.mail = mail
        self.last_seen = last_seen
        self.is_admin = is_admin
        self.crypto_key = crypto_key
        self.blacklist = blacklist or {
            "messages": [],
            "skills": [],
            "intents": []
        }


class JsonClientDatabase(JsonDatabase):
    def __init__(self, path=CONFIGURATION["database"]):
        super().__init__("clients", path)

    def update_timestamp(self, key, timestamp):
        user = self.get_client_by_api_key(key)
        if not user:
            return False
        item_id = self.get_item_id(user)
        user["last_seen"] = timestamp
        self.update_item(item_id, user)
        return True

    def delete_client(self, key):
        user = self.get_client_by_api_key(key)
        if user:
            item_id = self.get_item_id(user)
            self.update_item(item_id, Client(-1, api_key="revoked"))
            return True
        return False

    def change_key(self, old_key, new_key):
        user = self.get_client_by_api_key(old_key)
        if not user:
            return False
        item_id = self.get_item_id(user)
        user["api_key"] = new_key
        self.update_item(item_id, user)
        return True

    def change_crypto_key(self, api_key, new_key):
        user = self.get_client_by_api_key(api_key)
        if not user:
            return False
        item_id = self.get_item_id(user)
        user["crypto_key"] = new_key
        self.update_item(item_id, user)
        return True

    def get_crypto_key(self, api_key):
        user = self.get_client_by_api_key(api_key)
        if not user:
            return None
        return user["crypto_key"]

    def change_name(self, new_name, key):
        user = self.get_client_by_api_key(key)
        if not user:
            return False
        item_id = self.get_item_id(user)
        user["name"] = new_name
        self.update_item(item_id, user)
        return True

    def change_blacklist(self, blacklist, key):
        if isinstance(blacklist, dict):
            blacklist = json.dumps(blacklist)
        user = self.get_client_by_api_key(key)
        if not user:
            return False
        item_id = self.get_item_id(user)
        user["blacklist"] = blacklist
        self.update_item(item_id, user)
        return True

    def get_blacklist_by_api_key(self, api_key):
        search = self.search_by_value("api_key", api_key)
        if len(search):
            return search[0]["blacklist"]
        return None

    def get_client_by_api_key(self, api_key):
        search = self.search_by_value("api_key", api_key)
        if len(search):
            return search[0]
        return None

    def get_clients_by_name(self, name):
        return self.search_by_value("name", name)

    def add_client(self, name=None, mail=None, key="",
                   admin=None, blacklist=None, crypto_key="RESISTENCEisFUTILE"):

        user = self.get_client_by_api_key(key)
        item_id = self.get_item_id(user)
        if crypto_key is not None:
            crypto_key = crypto_key[:16]
        if item_id >= 0:
            if name:
                user["name"] = name
            if mail:
                user["mail"] = mail
            if blacklist:
                user["blacklist"] = blacklist
            if admin is not None:
                user["is_admin"] = admin
            user["crypto_key"] = crypto_key

            self.update_item(item_id, user)
        else:
            user = Client(api_key=key, name=name, mail=mail,
                          blacklist=blacklist, crypto_key=crypto_key,
                          client_id=self.total_clients() + 1,
                          is_admin=admin)
            self.add_item(user)

    def total_clients(self):
        return len(self)

    def __enter__(self):
        """ Context handler """
        return self

    def __exit__(self, _type, value, traceback):
        """ Commits changes and Closes the session """
        try:
            self.commit()
        except Exception as e:
            LOG.error(e)




from setuptools import setup

setup(
    name='jarbas_hive_mind',
    version='0.10.6',
    packages=['jarbas_hive_mind',
              'jarbas_hive_mind.master',
              'jarbas_hive_mind.slave',
              'jarbas_hive_mind.configuration',
              'jarbas_hive_mind.database',
              'jarbas_hive_mind.utils',
              'jarbas_hive_mind.discovery'],
    include_package_data=True,
    install_requires=["pyopenssl",
                      "service_identity",
                      "autobahn",
                      "twisted",
                      "ovos_utils>=0.0.1",
                      "json_database>=0.2.6",
                      "pycryptodome",
                      "upnpclient>=0.0.8"],
    url='https://github.com/JarbasAl/hive_mind',
    license='MIT',
    author='jarbasAI',
    author_email='jarbasai@mailfence.com',
    description='Mesh Networking utilities for mycroft core'
)



from http.server import BaseHTTPRequestHandler, HTTPServer
import threading

PORT_NUMBER = 8080


class UPNPHTTPServerHandler(BaseHTTPRequestHandler):
    """
    A HTTP handler that serves the UPnP XML files.
    """

    # Handler for the GET requests
    def do_GET(self):
        if self.path == "/" + self.server.scpd_xml_path:
            self.send_response(200)
            self.send_header('Content-type', 'application/xml')
            self.end_headers()
            self.wfile.write(self.scpd_xml.encode())
            return
        if self.path == "/" + self.server.device_xml_path:
            self.send_response(200)
            self.send_header('Content-type', 'application/xml')
            self.end_headers()
            self.wfile.write(self.device_xml.encode())
            return
        else:
            self.send_response(404)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            self.wfile.write(b"Not found.")
            return

    @property
    def services_xml(self):
        xml = """<serviceList>
            <service>
                <URLBase>{hive_url}</URLBase>
                <serviceType>urn:jarbasAi:HiveMind:service:Master</serviceType>
                <serviceId>urn:jarbasAi:HiveMind:serviceId:HiveMindNode</serviceId>
                <controlURL>/HiveMind</controlURL>
                <eventSubURL/>
                <SCPDURL>{scpd_path}</SCPDURL>
            </service>
        </serviceList>"""
        return xml.format(scpd_path=self.server.scpd_xml_path,
                          hive_url=self.server.presentation_url)

    @property
    def device_xml(self):
        """
        Get the main device descriptor xml file.
        """
        xml = """<root>
    <specVersion>
        <major>{major}</major>
        <minor>{minor}</minor>
    </specVersion>
    <device>
        <deviceType>urn:schemas-upnp-org:device:Basic:1</deviceType>
        <friendlyName>{friendly_name}</friendlyName>
        <manufacturer>{manufacturer}</manufacturer>
        <manufacturerURL>{manufacturer_url}</manufacturerURL>
        <modelDescription>{model_description}</modelDescription>
        <modelName>{model_name}</modelName>
        <modelNumber>{model_number}</modelNumber>
        <modelURL>{model_url}</modelURL>
        <serialNumber>{serial_number}</serialNumber>
        <UDN>uuid:{uuid}</UDN>
        {services_xml}
        <presentationURL>{presentation_url}</presentationURL>
    </device>
</root>"""
        return xml.format(friendly_name=self.server.friendly_name,
                          manufacturer=self.server.manufacturer,
                          manufacturer_url=self.server.manufacturer_url,
                          model_description=self.server.model_description,
                          model_name=self.server.model_name,
                          model_number=self.server.model_number,
                          model_url=self.server.model_url,
                          serial_number=self.server.serial_number,
                          uuid=self.server.uuid,
                          presentation_url=self.server.presentation_url,
                          scpd_path=self.server.scpd_xml_path,
                          services_xml=self.services_xml,
                          major=self.server.major_version,
                          minor=self.server.minor_version
                          # device_xml_path=self.device_xml_path
                          )

    @property
    def scpd_xml(self):
        """
        Get the device WSD file.
        """
        return """<scpd xmlns="urn:schemas-upnp-org:service-1-0">
<specVersion>
<major>1</major>
<minor>0</minor>
</specVersion>
</scpd>"""


class UPNPHTTPServerBase(HTTPServer):
    """
    A simple HTTP server that knows the information about a UPnP device.
    """

    def __init__(self, server_address, request_handler_class):
        HTTPServer.__init__(self, server_address, request_handler_class)
        self.port = None
        self.friendly_name = None
        self.manufacturer = None
        self.manufacturer_url = None
        self.model_description = None
        self.model_name = None
        self.model_url = None
        self.serial_number = None
        self.uuid = None
        self.presentation_url = None
        self.scpd_xml_path = None
        self.device_xml_path = None
        self.major_version = None
        self.minor_version = None


class UPNPHTTPServer(threading.Thread):
    """
    A thread that runs UPNPHTTPServerBase.
    """

    def __init__(self, port, friendly_name, manufacturer, manufacturer_url,
                 model_description, model_name,
                 model_number, model_url, serial_number, uuid,
                 presentation_url, host=""):
        threading.Thread.__init__(self, daemon=True)
        self.server = UPNPHTTPServerBase(('', port), UPNPHTTPServerHandler)
        self.server.port = port
        self.server.friendly_name = friendly_name
        self.server.manufacturer = manufacturer
        self.server.manufacturer_url = manufacturer_url
        self.server.model_description = model_description
        self.server.model_name = model_name
        self.server.model_number = model_number
        self.server.model_url = model_url
        self.server.serial_number = serial_number
        self.server.uuid = uuid
        self.server.presentation_url = presentation_url
        self.server.scpd_xml_path = 'scpd.xml'
        self.server.device_xml_path = "device.xml"
        self.server.major_version = 0
        self.server.minor_version = 1
        self.host = host

    @property
    def path(self):
        path = 'http://{ip}:{port}/{path}'.format(ip=self.host,
                                                  port=8088,
                                                  path=self.server.device_xml_path)
        return path

    def run(self):
        self.server.serve_forever()



# Licensed under the MIT license
# http://opensource.org/licenses/mit-license.php

# Copyright 2005, Tim Potter <tpot@samba.org>
# Copyright 2006 John-Mark Gurney <gurney_j@resnet.uroegon.edu>
# Copyright (C) 2006 Fluendo, S.A. (www.fluendo.com).
# Copyright 2006,2007,2008,2009 Frank Scholz <coherence@beebits.net>
# Copyright 2016 Erwan Martin <public@fzwte.net>
#
# Implementation of a SSDP server.
#

import random
import time
import socket
from email.utils import formatdate
from errno import ENOPROTOOPT
import threading
from ovos_utils.log import LOG

SSDP_PORT = 1900
SSDP_ADDR = '239.255.255.250'
SERVER_ID = 'HiveMind SSDP Server V0.1'


class SSDPServer(threading.Thread):
    """A class implementing a SSDP server.  The notify_received and
    searchReceived methods are called when the appropriate type of
    datagram is received by the server."""
    known = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sock = None
        self.running = False

    def run(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if hasattr(socket, "SO_REUSEPORT"):
            try:
                self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            except socket.error as le:
                # RHEL6 defines SO_REUSEPORT but it doesn't work
                if le.errno == ENOPROTOOPT:
                    pass
                else:
                    raise

        addr = socket.inet_aton(SSDP_ADDR)
        interface = socket.inet_aton('0.0.0.0')
        cmd = socket.IP_ADD_MEMBERSHIP
        self.sock.setsockopt(socket.IPPROTO_IP, cmd, addr + interface)
        self.sock.bind(('0.0.0.0', SSDP_PORT))
        self.sock.settimeout(1)
        self.running = True
        while self.running:
            try:
                data, addr = self.sock.recvfrom(1024)
                self.datagram_received(data, addr)
            except socket.timeout:
                continue
        self.shutdown()
        self.running = False

    def shutdown(self):
        for st in self.known:
            if self.known[st]['MANIFESTATION'] == 'local':
                self.do_byebye(st)

    def datagram_received(self, data, host_port):
        """Handle a received multicast datagram."""

        (host, port) = host_port

        try:
            header, payload = data.decode().split('\r\n\r\n')[:2]
        except ValueError as err:
            LOG.error(err)
            return

        lines = header.split('\r\n')
        cmd = lines[0].split(' ')
        lines = map(lambda x: x.replace(': ', ':', 1), lines[1:])
        lines = filter(lambda x: len(x) > 0, lines)

        headers = [x.split(':', 1) for x in lines]
        headers = dict(map(lambda x: (x[0].lower(), x[1]), headers))

        #LOG.info('SSDP command %s %s - from %s:%d' % (cmd[0], cmd[1], host,
        #                                            port))
        #LOG.debug('with headers: {}.'.format(headers))
        if cmd[0] == 'M-SEARCH' and cmd[1] == '*':
            # SSDP discovery
            self.discovery_request(headers, (host, port))
        elif cmd[0] == 'NOTIFY' and cmd[1] == '*':
            # SSDP presence
            # LOG.debug('NOTIFY *')
            pass
        else:
            LOG.warning('Unknown SSDP command %s %s' % (cmd[0], cmd[1]))

    def register(self, manifestation, usn, st, location, server=SERVER_ID, cache_control='max-age=1800', silent=False,
                 host=None):
        """Register a service or device that this SSDP server will
        respond to."""

        LOG.info('Registering %s (%s)' % (st, location))

        self.known[usn] = {}
        self.known[usn]['USN'] = usn
        self.known[usn]['LOCATION'] = location
        self.known[usn]['ST'] = st
        self.known[usn]['EXT'] = ''
        self.known[usn]['SERVER'] = server
        self.known[usn]['CACHE-CONTROL'] = cache_control

        self.known[usn]['MANIFESTATION'] = manifestation
        self.known[usn]['SILENT'] = silent
        self.known[usn]['HOST'] = host
        self.known[usn]['last-seen'] = time.time()

        if manifestation == 'local' and self.sock:
            self.do_notify(usn)

    def unregister(self, usn):
        LOG.info("Un-registering %s" % usn)
        del self.known[usn]

    def is_known(self, usn):
        return usn in self.known

    def send_it(self, response, destination, delay, usn):
        #LOG.debug('send discovery response delayed by %ds for %s to %r' % (
        #    delay, usn, destination))
        try:
            self.sock.sendto(response.encode(), destination)
        except (AttributeError, socket.error) as msg:
            LOG.warning("failure sending out byebye notification: %r" % msg)

    def discovery_request(self, headers, host_port):
        """Process a discovery request.  The response must be sent to
        the address specified by (host, port)."""

        (host, port) = host_port

        #LOG.info('Discovery request from (%s,%d) for %s' % (host, port,
        #                                                  headers['st']))
        #LOG.info('Discovery request for %s' % headers['st'])

        # Do we know about this service?
        for i in self.known.values():
            if i['MANIFESTATION'] == 'remote':
                continue
            if headers['st'] == 'ssdp:all' and i['SILENT']:
                continue
            if i['ST'] == headers['st'] or headers['st'] == 'ssdp:all':
                response = ['HTTP/1.1 200 OK']

                usn = None
                for k, v in i.items():
                    if k == 'USN':
                        usn = v
                    if k not in ('MANIFESTATION', 'SILENT', 'HOST'):
                        response.append('%s: %s' % (k, v))

                if usn:
                    response.append('DATE: %s' % formatdate(timeval=None, localtime=False, usegmt=True))

                    response.extend(('', ''))
                    delay = random.randint(0, int(headers['mx']))

                    self.send_it('\r\n'.join(response), (host, port), delay, usn)

    def do_notify(self, usn):
        """Do notification"""

        if self.known[usn]['SILENT']:
            return
        LOG.info('Sending alive notification for %s' % usn)

        resp = [
            'NOTIFY * HTTP/1.1',
            'HOST: %s:%d' % (SSDP_ADDR, SSDP_PORT),
            'NTS: ssdp:alive',
        ]
        stcpy = dict(self.known[usn].items())
        stcpy['NT'] = stcpy['ST']
        del stcpy['ST']
        del stcpy['MANIFESTATION']
        del stcpy['SILENT']
        del stcpy['HOST']
        del stcpy['last-seen']

        resp.extend(map(lambda x: ': '.join(x), stcpy.items()))
        resp.extend(('', ''))
        LOG.debug('do_notify content', resp)
        try:
            self.sock.sendto('\r\n'.join(resp).encode(), (SSDP_ADDR, SSDP_PORT))
            self.sock.sendto('\r\n'.join(resp).encode(), (SSDP_ADDR, SSDP_PORT))
        except (AttributeError, socket.error) as msg:
            LOG.warning("failure sending out alive notification: %r" % msg)

    def do_byebye(self, usn):
        """Do byebye"""

        LOG.info('Sending byebye notification for %s' % usn)

        resp = [
            'NOTIFY * HTTP/1.1',
            'HOST: %s:%d' % (SSDP_ADDR, SSDP_PORT),
            'NTS: ssdp:byebye',
        ]
        try:
            stcpy = dict(self.known[usn].items())
            stcpy['NT'] = stcpy['ST']
            del stcpy['ST']
            del stcpy['MANIFESTATION']
            del stcpy['SILENT']
            del stcpy['HOST']
            del stcpy['last-seen']
            resp.extend(map(lambda x: ': '.join(x), stcpy.items()))
            resp.extend(('', ''))
            #LOG.debug('do_byebye content', resp)
            if self.sock:
                try:
                    self.sock.sendto('\r\n'.join(resp), (SSDP_ADDR, SSDP_PORT))
                except (AttributeError, socket.error) as msg:
                    LOG.error("failure sending out byebye notification: %r" %
                            msg)
        except KeyError as msg:
            LOG.error("error building byebye notification: %r" % msg)


from twisted.internet.error import DNSLookupError


class HiveMindException(Exception):
    """ An Exception inside the HiveMind"""


class UnauthorizedKeyError(HiveMindException):
    """ Invalid Key provided """


class WrongEncryptionKey(HiveMindException):
    """ Wrong Encryption Key"""


class DecryptionKeyError(WrongEncryptionKey):
    """ Could not decrypt payload """


class EncryptionKeyError(WrongEncryptionKey):
    """ Could not encrypt payload """


class ConnectionError(HiveMindException):
    """ Could not connect to the HiveMind"""


class SecureConnectionFailed(ConnectionError):
    """ Could not connect by SSL """


class HiveMindEntryPointNotFound(ConnectionError, DNSLookupError):
    """ can not connect to provided address """
